import sys
import cd_solver
import numpy as np
norm = np.linalg.norm
import scipy.sparse as sp
import copy
from sklearn.datasets import load_svmlight_file, load_iris
import matplotlib.pyplot as plt
import time

# get data
plt.close('all')

do_toy_problem = 0
if do_toy_problem:
    # min c.dot(x) : x >= 0, Mx <= b
    d = 3
    n = 4
    M = sp.csc_matrix(np.array([[2,4,5,7], [1,1,2,2], [1,2,3,3]]), dtype='float')
    c = -np.array([7,9,18,17])
    b = np.array([41,17,24])

    prob_toy = cd_solver.Problem(N=M.shape[1], f=['linear'], Af=c, g=['ineq_const']*n,
                                 blocks=np.arange(M.shape[1]+1),
                                 h=['ineq_const']*d, Ah=-M, bh=-b,
                                 blocks_h=np.arange(M.shape[0]+1))

    prob_toy2 = copy.copy(prob_toy)
    prob_toy3 = copy.copy(prob_toy)
    cd_solver.coordinate_descent(prob_toy, algorithm='pure-cd', verbose=0.01,
                                 print_style='smoothed_gap', average=True,
                                 restart_period=10,
                                 fixed_restart_period='adaptive',
                                 show_restart_info=True, 
                                 max_iter=10000,
                                 adaptive_steps='fercoq_stochastic_autoregressive', step_size_factor=1)
    cd_solver.coordinate_descent(prob_toy2, algorithm='pure-cd', verbose=0.01,
                                 print_style='smoothed_gap', average=True,
                                 restart_period=10,
                                 fixed_restart_period='adaptive',
                                 show_restart_info=False,
                                 max_iter=10000,
                                 step_size_factor=1)
    cd_solver.coordinate_descent(prob_toy3, algorithm='pure-cd', verbose=0.01,
                                 print_style='smoothed_gap', average=True,
                                 restart_period=200,
                                 fixed_restart_period='fixed',
                                 show_restart_info=False, 
                                 max_iter=10000,
                                 step_size_factor=1)

    prints = np.array(prob_toy.printed_values)
    prints2 = np.array(prob_toy2.printed_values)
    prints3 = np.array(prob_toy3.printed_values)
    plt.semilogy(prints[:,1], np.max(prints[:, 4:7], axis=1),
                 prints2[:,1], np.max(prints2[:, 4:7], axis=1),
                 prints3[:,1], np.max(prints3[:, 4:7], axis=1))
    plt.legend(['adaptive restart and adaptive steps', 'adaptive restart', 'fixed restart 200'])
    plt.show()

do_toy_problem_quadratic = False
if do_toy_problem_quadratic:
    # min_x mu/2 x**2 st ax = b
    a = 1
    b = 3
    mu = 0.01
    prob_toy_quad = cd_solver.Problem(N=1, f=['square'], Af=[np.sqrt(mu/2)],
                                 h=['eq_const'], Ah=np.array([[a]]), bh=[-b],
                                 blocks_h=[0, 1])
    
    cd_solver.coordinate_descent(prob_toy_quad, algorithm='pure-cd',
                                 verbose=0.01,
                                 print_style='smoothed_gap', average=True,
                                 restart_period=2,
                                 fixed_restart_period='adaptive',
                                 show_restart_info=True, 
                                 max_iter=2000,
                                 adaptive_steps='fercoq_stochastic_autoregressive', step_size_factor=0.1)

do_sparse_SVM = False
if do_sparse_SVM:
    # sparse SVM
    # min_w C sum_i max(0, 1 - y_i(x_i w)) + ||w||_1
    dataset = 'a1a'
    (X, y) = load_svmlight_file(dataset + '.data')

    #data = load_iris()
    #y = data.target
    #y[y==0] = -1
    #y[y==2] = -1
    #X = sp.csc_matrix(data.data)

    X = sp.csc_matrix(X.multiply(1./np.sqrt(X.multiply(X).sum(axis=0))))
    Xy = (X.T.multiply(y)).T
    lamb = 1


    # define problem
    prob = cd_solver.Problem(N=Xy.shape[1], f=['linear'], Af=-Xy.sum(axis=0),
                             cf=[0.5], bf=[-Xy.shape[0]],
                             g=['abs']*Xy.shape[1], cg=[lamb]*Xy.shape[1],
                             h=['abs']*Xy.shape[0], Ah=-Xy, ch=[0.5]*Xy.shape[0],
                             bh=[-1]*Xy.shape[0])
    prob2 = copy.copy(prob)

    prob3 = copy.copy(prob)
    max_iter = 100000

    # run pure-cd with averaging and restart
    print('pure-cd with averaging and restart on '+dataset)
    cd_solver.coordinate_descent(prob2, algorithm='pure-cd', average=True, restart_period=10, verbose=0.1, print_style='smoothed_gap', max_iter=20000, tolerance=1e-9, fixed_restart_period='adaptive')

    # run pure-cd
    print('pure-cd on '+dataset)
    cd_solver.coordinate_descent(prob, algorithm='pure-cd', average=False, verbose=0.1, print_style='smoothed_gap', max_iter=20000, tolerance=1e-9)

    # run pure-cd with adaptive steps
    print('pure-cd on '+dataset)
    cd_solver.coordinate_descent(prob3, algorithm='pure-cd', average=False, verbose=0.1, print_style='smoothed_gap', max_iter=20000, tolerance=1e-9, adaptive_steps='fercoq_stochastic_autoregressive')

    prints = np.array(prob.printed_values)
    plt.semilogy(prints[:,1], np.max(prints[:, 4:7], axis=1))
    prints = np.array(prob2.printed_values)
    plt.semilogy(prints[:,1], np.max(prints[:, 4:7], axis=1), linestyle='--')
    prints = np.array(prob3.printed_values)
    plt.semilogy(prints[:,1], np.max(prints[:, 4:7], axis=1), linestyle='-.')
    plt.xlabel('iteration')
    plt.ylabel('max(smoothed gap, dual infeasibility)')
    plt.legend(['Pure-cd', 'RA-Pure-cd', 'Pure-cd-adaptive-steps'])
    plt.show()


# run other algorithms
do_ridge = False
if do_ridge:
    # ridge regression on real-sim
    (X, y) = load_svmlight_file('a1a.data')
    # X2 = sp.csc_matrix(X.multiply(1./np.sqrt(X.multiply(X).sum(axis=0))))
    a = 0.001 / X.shape[0]
    n = X.shape[0]
    d = X.shape[1]
    prob_ridge = cd_solver.Problem(N=X.shape[1], h=['square']*n, Ah=X, bh=y, ch=[0.5]*n,
                                   g=['square']*d, cg=[a/2]*d)

    prob_ridge2 = copy.copy(prob_ridge)
    prob_ridge4 = copy.copy(prob_ridge)
    prob_ridge5 = copy.copy(prob_ridge)
    prob_ridge6 = copy.copy(prob_ridge)
    prob_ridge7 = copy.copy(prob_ridge)

    prob_ridge3 = cd_solver.Problem(N=X.shape[1], f=['square'], Af=X, bf=y, blocks_f=[0, X.shape[0]], cf=[0.5],
                                    g=['square']*d, cg=[a/2]*d)
    
#    cd_solver.coordinate_descent(prob_ridge7, algorithm='pure-cd', verbose=0.3, tolerance=1e-9, print_style='smoothed_gap', step_size_factor=500, average=True,
#                                 restart_period=1, fixed_restart_period='just_check', adaptive_steps='fercoq_residual_norm', max_iter=10000, show_restart_info=False)

    verbose_time = 0.1
    max_iter = 3000
    tol = 1e-5

    print("fercoq stochastic lognormal")
    cd_solver.coordinate_descent(prob_ridge4, algorithm='pure-cd', verbose=verbose_time, tolerance=tol, print_style='smoothed_gap', step_size_factor=500, average=True,
                                 restart_period=10, fixed_restart_period='adaptive', adaptive_steps="fercoq_stochastic_lognormal", max_iter=max_iter, show_restart_info=False)

    print("fercoq stochastic ar")
    cd_solver.coordinate_descent(prob_ridge5, algorithm='pure-cd', verbose=verbose_time, tolerance=tol, print_style='smoothed_gap', step_size_factor=500, average=True,
                                 restart_period=10, fixed_restart_period='adaptive', adaptive_steps='fercoq_stochastic_autoregressive', max_iter=max_iter, show_restart_info=False)

    print("goldstein for pure-cd")
    cd_solver.coordinate_descent(prob_ridge6, algorithm='pure-cd', verbose=verbose_time, tolerance=tol, print_style='smoothed_gap', step_size_factor=500, average=True,
                                 restart_period=10, fixed_restart_period='adaptive', adaptive_steps='goldstein', max_iter=max_iter, show_restart_info=False)

    print("rapdhg")
    cd_solver.coordinate_descent(prob_ridge2, algorithm='pure-cd', verbose=verbose_time, tolerance=tol, print_style='smoothed_gap', step_size_factor=500, average=True,
                                 restart_period=10, fixed_restart_period='adaptive', max_iter=max_iter)
    if 1:
        print("pdhg")
        cd_solver.coordinate_descent(prob_ridge, algorithm='pure-cd', verbose=verbose_time, tolerance=tol, print_style='smoothed_gap', step_size_factor=500, max_iter=max_iter)

        print("approx")
        cd_solver.coordinate_descent(prob_ridge3, algorithm='smart-cd', verbose=verbose_time, tolerance=tol, print_style='smoothed_gap', restart_period=np.ceil(2*np.exp(1/2.)/np.sqrt(a)*sp.linalg.norm(X))-1, max_iter=max_iter, fixed_restart_period=True)



    prints = np.array(prob_ridge.printed_values)
    plt.semilogy(prints[:,1], np.max(prints[:, 4:7], axis=1))
    prints = np.array(prob_ridge2.printed_values)
    plt.semilogy(prints[:,1], np.max(prints[:, 4:7], axis=1), linestyle=':')
    prints = np.array(prob_ridge3.printed_values)
    plt.semilogy(prints[:,1], np.max(prints[:, 4:7], axis=1), linestyle=(0, (3, 5, 1, 5)))
    prints = np.array(prob_ridge4.printed_values)
    plt.semilogy(prints[:,1], np.max(prints[:, 4:7], axis=1), linestyle='--')
    prints = np.array(prob_ridge5.printed_values)
    plt.semilogy(prints[:,1], np.max(prints[:, 4:7], axis=1), linestyle='-')
    prints = np.array(prob_ridge6.printed_values)
    plt.semilogy(prints[:,1], np.max(prints[:, 4:7], axis=1), linestyle='-.')
    plt.xlabel('iteration')
    plt.ylabel('smoothed duality gap')
    plt.legend(['pure-cd', 'RA-pure-cd', 'R-approx', 'lognormal-pure-cd', 'autoregressive-pure-cd', 'Goldstein-pure-cd'])
    plt.show()

    history_adaptive_steps = np.array(prob_ridge5.history_adaptive_steps)
    taus = history_adaptive_steps[:,0]
    rhos = history_adaptive_steps[:,1]
    plt.plot(np.log(taus), -np.log(1-rhos))
    plt.show()

do_l1_rof = False
if do_l1_rof:
    # L1 ROF
    # min_x ||Dx||_{2,1} + a ||x - im||_1
    im = plt.imread('cameraman.png')
    # im = im[:30,:30]
    N = np.prod(im.shape)
    # plt.imshow(im, cmap='gray')
    Dx = sp.diags([-np.ones(im.shape[0]), np.ones(im.shape[0])], offsets=[0, 1])
    Dx = sp.csc_matrix(Dx)[:-1]  # the last line is not a discrete gradient
    Dy = sp.diags([-np.ones(im.shape[1]), np.ones(im.shape[1])], offsets=[0, 1])
    Dy = sp.csc_matrix(Dy)[:-1]
    Dx = sp.kron(Dx, sp.eye(im.shape[1]))
    Dy = sp.kron(sp.eye(im.shape[0]), Dy)
    twoDgradient = sp.vstack([Dx, Dy], format='csc')
    twoDgradient.eliminate_zeros()
    # reorder the matrix
    twoDgradient = sp.csc_matrix((twoDgradient.data, 2 * (twoDgradient.indices % N) + twoDgradient.indices // N, twoDgradient.indptr), (2*N, N))
    a = 1.9
    maxiter = 5e4

    skip_pdhg = False
    if skip_pdhg == False:
        #pb_l1_rof = cd_solver.Problem(N=N, g=["abs"], cg=[a], bg=im.ravel(),
        #            blocks=[0, N],
        #            h=["norm2"] * N, blocks_h=np.arange(0, 2*N + 1, 2),
        #                          Ah=twoDgradient, x_init=im.ravel())

        #cd_solver.coordinate_descent(pb_l1_rof, algorithm='pure-cd', verbose=100, tolerance=1e-6, print_style='smoothed_gap', max_iter=1e6, step_size_factor=1, max_time=1e4, average=False)

        pb_l1_rof = cd_solver.Problem(N=N, g=["abs"], cg=[a], bg=im.ravel(),
                    blocks=[0, N],
                    h=["norm2"] * N, blocks_h=np.arange(0, 2*N + 1, 2),
                                   Ah=twoDgradient, x_init=im.ravel())

        print('run rapdhg on cameraman l1_rof with adaptive step sizes')
        cd_solver.coordinate_descent(pb_l1_rof, algorithm='pure-cd', verbose=10, tolerance=1e-6, print_style='smoothed_gap', average=False, step_size_factor=1, restart_period=1000, fixed_restart_period='just_check', max_iter=maxiter, max_time=1e3, min_change_in_x=0, show_restart_info=True)

    pb_l1_rof_cd = cd_solver.Problem(N=N, g=["abs"]*N, cg=[a]*N, bg=im.ravel(),
                                     h=["norm2"] * N, blocks_h=np.arange(0, 2*N + 1, 2),
                                     Ah=twoDgradient, x_init=im.ravel())

    pb_l1_rof_cd2 = copy.copy(pb_l1_rof_cd)
    pb_l1_rof_cd3 = copy.copy(pb_l1_rof_cd)
    pb_l1_rof_cd4 = copy.copy(pb_l1_rof_cd)

    
    cd_solver.coordinate_descent(pb_l1_rof_cd, algorithm='pure-cd', verbose=10, tolerance=1e-6, print_style='smoothed_gap', average=False, step_size_factor=1, restart_period=1000, fixed_restart_period='just_check', max_iter=maxiter, max_time=1e3, min_change_in_x=0, show_restart_info=True)

    cd_solver.coordinate_descent(pb_l1_rof_cd2, algorithm='pure-cd', verbose=10, tolerance=1e-6, print_style='smoothed_gap', average=False, step_size_factor=1, restart_period=1000, fixed_restart_period='just_check', max_iter=maxiter, max_time=1e3, min_change_in_x=0, show_restart_info=True, adaptive_steps='goldstein')
    
    cd_solver.coordinate_descent(pb_l1_rof_cd3, algorithm='pure-cd', verbose=10, tolerance=1e-6, print_style='smoothed_gap', average=False, step_size_factor=1, restart_period=1000, fixed_restart_period='just_check', max_iter=maxiter, max_time=1e3, min_change_in_x=0, show_restart_info=True, adaptive_steps='fercoq_stochastic_lognormal_no_warmup')

    cd_solver.coordinate_descent(pb_l1_rof_cd4, algorithm='pure-cd', verbose=10, tolerance=1e-6, print_style='smoothed_gap', average=False, step_size_factor=1, restart_period=1000, fixed_restart_period='just_check', max_iter=maxiter, max_time=1e3, min_change_in_x=0, show_restart_info=True, adaptive_steps='fercoq_stochastic_lognormal')

    
    def norm21(z):
        z = z.reshape((-1,2))
        return np.sum(np.sqrt(np.sum(z**2, axis=1)))

    def norm2inf(z):
        z = z.reshape((-1,2))
        return np.max(np.sqrt(np.sum(z**2, axis=1)))

    def value(x):
        return norm21(twoDgradient * x) + a * norm(x - im.ravel(), 1)

    def st(x, a):
        return np.sign(x) * np.maximum(0, np.abs(x) - a)

    def dual_value(y, beta, x_center):
        if norm2inf(y) > 1.00001:
            return np.inf
        DTy = twoDgradient.T.dot(y)
        dval = im.ravel().dot(DTy)
        x_bar = st(x_center + DTy/beta, a/beta)
        dval += DTy.dot(x_bar) - a*norm(x_bar, 1) - beta/2*norm(x_bar - x_center,2)**2
        return dval

    #np.save('pb_l1_rof_solution', pb_l1_rof2.sol.reshape(im.shape))

    #if skip_pdhg == True:
    #    prints = np.load('prints_pb_l1_rof.npy')
    #else:
    #    prints = np.array(pb_l1_rof.printed_values)
    #    np.save('prints_pb_l1_rof', prints)

    prints = np.array(pb_l1_rof_cd.printed_values)
    plt.semilogy(prints[:,0], np.max(prints[:, 4:7], axis=1), linestyle=':')
    prints = np.array(pb_l1_rof_cd2.printed_values)
    plt.semilogy(prints[:,0], np.max(prints[:, 4:7], axis=1), linestyle='--')
    prints = np.array(pb_l1_rof_cd3.printed_values)
    plt.semilogy(prints[:,0], np.max(prints[:, 4:7], axis=1), linestyle='-')
    prints = np.array(pb_l1_rof_cd4.printed_values)
    plt.semilogy(prints[:,0], np.max(prints[:, 4:7], axis=1), linestyle='-.')

    plt.xlabel('time (s)')
    plt.ylabel('max(smoothed gap, dual infeasibility)')
    plt.legend(['pure-cd', 'pure-cd + residual balance', 'pure-cd + iid convergence monitoring', 'pure-cd + residual balance + iid convergence monitoring'])
    plt.show()

    plt.imshow(pb_l1_rof_cd2.sol.reshape(im.shape), cmap='gray'); plt.show()
    
"""
# test moyennes

sols = []
dual_sols = []
for k in range(5):
    cd_solver.coordinate_descent(problem, algorithm='pure-cd', average=False,
                             verbose=10, print_style='smoothed_gap',
                             max_iter=k+1, tolerance=1e-6)

    sols.append(problem.sol)
    dual_sols.append(problem.dual_sol)

cd_solver.coordinate_descent(problem, algorithm='pure-cd', average=True,
                             restart_period=5,
                             verbose=10, print_style='smoothed_gap',
                             max_iter=5, tolerance=1e-6)
"""

do_sqrt_lasso = True
if do_sqrt_lasso:
    
    dataset = 'leukemia'
    (X, y) = load_svmlight_file(dataset)
    X = sp.csc_matrix(X - X.mean(axis=0))
    X = sp.csc_matrix(X.multiply(1./np.sqrt(X.multiply(X).sum(axis=0))))
    lambda_max = norm(X.T @ y, np.inf) / norm(y, 2)
    lamb = lambda_max / 1.1
    N = X.shape[1]
    
    prob_sqrt_lasso = cd_solver.Problem(N=N, h=['norm2'], Ah=X,
                             ch=[1], bh=y,
                             g=['abs']*N, cg=[lamb]*N,
                             blocks=np.arange(N+1),
                             blocks_h=[0, X.shape[0]])
    prob_sqrt_lasso2 = copy.copy(prob_sqrt_lasso)
    prob_sqrt_lasso3 = copy.copy(prob_sqrt_lasso)

    prob_sqrt_lasso_full = cd_solver.Problem(N=N, h=['norm2'], Ah=X,
                             ch=[1], bh=y,
                             g=['abs'], cg=[lamb],
                             blocks=[0, N],
                             blocks_h=[0, X.shape[0]])
    prob_sqrt_lasso_full2 = copy.copy(prob_sqrt_lasso_full)


    cd_solver.coordinate_descent(prob_sqrt_lasso_full, algorithm='pure-cd',
                                 verbose=1, tolerance=1e-8,
                                 print_style='smoothed_gap', average=True,
                                 step_size_factor=1, restart_period=10,
                                 fixed_restart_period='adaptive', max_iter=1e5,
                                 max_time=1e2, min_change_in_x=0,
                                 show_restart_info=False,
                                 adaptive_steps=False, gamma_step_size=0.999)
    cd_solver.coordinate_descent(prob_sqrt_lasso_full2, algorithm='pure-cd',
                                 verbose=1, tolerance=1e-8,
                                 print_style='smoothed_gap', average=True,
                                 step_size_factor=1, restart_period=10,
                                 fixed_restart_period='adaptive', max_iter=1e5,
                                 max_time=1e2, min_change_in_x=0,
                                 show_restart_info=False,
                                 adaptive_steps='fercoq_residual_norm', gamma_step_size=0.999)

    cd_solver.coordinate_descent(prob_sqrt_lasso, algorithm='pure-cd',
                                 verbose=1, tolerance=1e-8,
                                 print_style='smoothed_gap', average=True,
                                 step_size_factor=1, restart_period=10,
                                 fixed_restart_period='adaptive', max_iter=3e5,
                                 max_time=1e2, min_change_in_x=0,
                                 show_restart_info=False,
                                 adaptive_steps=False, gamma_step_size=0.999)
    
    cd_solver.coordinate_descent(prob_sqrt_lasso2, algorithm='pure-cd',
                                 verbose=1, tolerance=1e-8,
                                 print_style='smoothed_gap', average=True,
                                 step_size_factor=1, restart_period=10,
                                 fixed_restart_period='adaptive', max_iter=1e5,
                                 max_time=1e2, min_change_in_x=0,
                                 show_restart_info=False,
                                 adaptive_steps='fercoq_stochastic_lognormal', gamma_step_size=0.999)

    cd_solver.coordinate_descent(prob_sqrt_lasso3, algorithm='pure-cd',
                                 verbose=1, tolerance=1e-8,
                                 print_style='smoothed_gap', average=True,
                                 step_size_factor=1, restart_period=10,
                                 fixed_restart_period='adaptive', max_iter=1e5,
                                 max_time=1e2, min_change_in_x=0,
                                 show_restart_info=False,
                                 adaptive_steps='fercoq_stochastic_lognormal_no_warmup', gamma_step_size=0.999)
    
    prints = np.array(prob_sqrt_lasso.printed_values)
    plt.semilogy(prints[:,0], np.max(prints[:, 4:7], axis=1), ':')
    prints = np.array(prob_sqrt_lasso2.printed_values)
    plt.semilogy(prints[:,0], np.max(prints[:, 4:7], axis=1), linestyle='--')
    prints = np.array(prob_sqrt_lasso3.printed_values)
    plt.semilogy(prints[:,0], np.max(prints[:, 4:7], axis=1), linestyle='-')
    prints = np.array(prob_sqrt_lasso_full.printed_values)
    plt.semilogy(prints[:,0], np.max(prints[:, 4:7], axis=1), linestyle='-.')
    prints = np.array(prob_sqrt_lasso_full2.printed_values)
    plt.semilogy(prints[:,0], np.max(prints[:, 4:7], axis=1), linestyle=(0, (10, 3)))
    plt.legend(['Pure-cd', 'Pure-cd + iid cm + residual balance', 'Pure-cd + iid convergence monitoring', 'PDHG', 'PDHG + convergence monitoring'])
    plt.xlabel('time (s)')
    plt.ylabel('max(smoothed gap, infeasibility)')
    plt.show()


do_tv_denoising = False
if do_tv_denoising == True:
    # min_x a ||Dx||_{2,1} + 0.5 ||x - im||_2^2
    im = plt.imread('poule_susie_noir_et_blanc.png')
    im = im[:, :650]
    # im = im[400:500, 0:100]
    N = np.prod(im.shape)
    # plt.imshow(im, cmap='gray')
    Dx = sp.diags([-np.ones(im.shape[0]), np.ones(im.shape[0])], offsets=[0, 1])
    Dx = sp.csc_matrix(Dx)[:-1]  # the last line is not a discrete gradient
    Dy = sp.diags([-np.ones(im.shape[1]), np.ones(im.shape[1])], offsets=[0, 1])
    Dy = sp.csc_matrix(Dy)[:-1]
    Dx = sp.kron(Dx, sp.eye(im.shape[1]))
    Dy = sp.kron(sp.eye(im.shape[0]), Dy)
    twoDgradient = sp.vstack([Dx, Dy], format='csc')
    twoDgradient.eliminate_zeros()
    # reorder the matrix
    twoDgradient = sp.csc_matrix((twoDgradient.data, 2 * (twoDgradient.indices % N) + twoDgradient.indices // N, twoDgradient.indptr), (2*N, N))
    a = 5e-3
    def norm2inf(z):
        z = z.reshape((-1,2))
        return np.max(np.sqrt(np.sum(z**2, axis=1)))

    maxtime = 3600
    maxiter = 500
    print_period = 10
    tol = 1e-10
    y_init = twoDgradient @ im.ravel()
    for i in range(N):
        y_init[2*i:2*i+2] *= a / np.max([np.linalg.norm(y_init[2*i:2*i+2]), 1e-5])
    
    
    pb_l2_rof_pure_cd = cd_solver.Problem(N=N, g=["square"]*N, cg=[0.5]*N, bg=im.ravel(),
                            blocks=np.arange(N+1),
                            h=["norm2"]*N, ch=[a]*N,
                            blocks_h=np.arange(0, 2*N + 1, 2),
                            Ah=twoDgradient, x_init=im.ravel(),
                            y_init=y_init
    )
    pb_l2_rof_pure_cd_residual_balance = copy.copy(pb_l2_rof_pure_cd)
    pb_l2_rof_pure_cd_stochastic_lognormal_no_warmup = copy.copy(pb_l2_rof_pure_cd)
    pb_l2_rof_pure_cd_autoregressive = copy.copy(pb_l2_rof_pure_cd)
    pb_l2_rof_pure_cd_stochastic_lognormal = copy.copy(pb_l2_rof_pure_cd)

    fixed_restart_period = 'just_check'
    
    cd_solver.coordinate_descent(pb_l2_rof_pure_cd, algorithm='pure-cd',
                                 verbose=print_period,
                                 tolerance=tol, print_style='smoothed_gap',
                                 max_iter=maxiter, step_size_factor=1,
                                 max_time=maxtime, average=True, restart_period=10,
                                 fixed_restart_period=fixed_restart_period)
    
    cd_solver.coordinate_descent(pb_l2_rof_pure_cd_autoregressive, algorithm='pure-cd',
                                 verbose=print_period,
                                 tolerance=tol, print_style='smoothed_gap',
                                 max_iter=maxiter, step_size_factor=1,
                                 max_time=maxtime, average=True, restart_period=10,
                                 fixed_restart_period=fixed_restart_period,
                                 adaptive_steps="fercoq_stochastic_autoregressive")
    cd_solver.coordinate_descent(pb_l2_rof_pure_cd_stochastic_lognormal, algorithm='pure-cd',
                                 verbose=print_period,
                                 tolerance=tol, print_style='smoothed_gap',
                                 max_iter=maxiter, step_size_factor=1,
                                 max_time=maxtime, average=True, restart_period=10,
                                 fixed_restart_period=fixed_restart_period,
                                 adaptive_steps="fercoq_stochastic_lognormal")
    
    cd_solver.coordinate_descent(pb_l2_rof_pure_cd_stochastic_lognormal_no_warmup, algorithm='pure-cd',
                                 verbose=print_period,
                                 tolerance=tol, print_style='smoothed_gap',
                                 max_iter=maxiter, step_size_factor=1,
                                 max_time=maxtime, average=True, restart_period=10,
                                 fixed_restart_period=fixed_restart_period,
                                 adaptive_steps="fercoq_stochastic_lognormal_no_warmup")

    cd_solver.coordinate_descent(pb_l2_rof_pure_cd_residual_balance, algorithm='pure-cd',
                                 verbose=print_period,
                                 tolerance=tol, print_style='smoothed_gap',
                                 max_iter=maxiter, step_size_factor=1,
                                 max_time=maxtime, average=True, restart_period=10,
                                 fixed_restart_period=fixed_restart_period,
                                 adaptive_steps="goldstein")
    
    plt.figure(1)
    #prints = np.array(pb_l2_rof_pdhg.printed_values)
    #plt.semilogy(prints[:,1], np.max(prints[:, 4:7], axis=1), ':')
    prints = np.array(pb_l2_rof_pure_cd.printed_values)
    plt.semilogy(prints[:,1], np.max(prints[:, 4:7], axis=1), ':')
    prints = np.array(pb_l2_rof_pure_cd_residual_balance.printed_values)
    plt.semilogy(prints[:,1], np.max(prints[:, 4:7], axis=1), '--')
    prints = np.array(pb_l2_rof_pure_cd_stochastic_lognormal_no_warmup.printed_values)
    plt.semilogy(prints[:,1], np.max(prints[:, 4:7], axis=1))
    prints = np.array(pb_l2_rof_pure_cd_stochastic_lognormal.printed_values)
    plt.semilogy(prints[:,1], np.max(prints[:, 4:7], axis=1), '-.')
    prints = np.array(pb_l2_rof_pure_cd_autoregressive.printed_values)
    plt.semilogy(prints[:,1], np.max(prints[:, 4:7], axis=1), linestyle=(0, (10, 3)))

    plt.xlabel('Iteration')
    plt.ylabel('duality gap')
    plt.legend(['Pure-cd', 'Pure-cd + residual balance', 'Pure-cd + iid convergence monitoring', 'Pure-cd + iid cm + residual balance', 'Pure-cd + AR1 cm + residual balance'])
    plt.show()

    memory = pb_l2_rof_pure_cd_residual_balance.history_adaptive_steps
    plt.figure(2)
    plt.semilogy(np.linspace(0, prints[-1,1], len(memory[7])), np.array(memory[7]))
    plt.show()

    plt.figure(3)
    plt.semilogy(np.linspace(0, prints[-1,1], len(memory[7])), 1-np.array(memory[2]))
    plt.show()
    
    plt.figure(4)
    plt.semilogy(np.linspace(0, prints[-1,1], len(memory[7])),
             np.array(memory[1]))
    plt.show()
