import numpy as np
import scipy.sparse as sp
import scipy.sparse.linalg as spl
import scipy.linalg
import scipy.special as sps
import scipy.optimize
sqrtm = scipy.linalg.sqrtm
norm = np.linalg.norm
inv = np.linalg.inv
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import pdb

plt.close('all')
eta = 0.001
mup = 0.01
mud = 0.1
Nx = 20
Ny = Nx-1

A = sp.csc_matrix(sp.diags([np.array([1] +  [1+eta] * (Nx-1)), -np.ones(Nx)], [0, -1])).toarray()
A = A[1:]
Q = mup * np.eye(Nx)
S = mud * np.eye(Ny)

b = np.array([1] + [0]*(Ny-1))
c = np.zeros(Nx)

gamma = 0.95

M1 = np.vstack([np.hstack([Q, A.T]),
                np.hstack([A, -S])])
HessL = M1

Ix = np.eye(Nx)
Iy = np.eye(Ny)


def L(x, y):
    return 0.5 * x @ Q @ x + c @ x + y @ (A @ x - b) - 0.5 * y @ S @ y

def smoothed_gap(x, y, beta=1):
    # implementation for Q being a multiple of the identity matrix
    # max_x' max_y' L(x,y') - L(x', y) - beta/2 (z - z')**2
    # 0 = -Qx' - beta (x' - x) - A.T y
    # 0 = -Sy' + Ax - b - beta (y' - y)
    px = 1 / (beta + Q[0,0]) * (beta * x - A.T @ y - c)
    py = 1 / (beta + S[0,0]) * (beta * y + A @ x - b)
    # return L(x, py) - L(px, y) - beta/2 * (norm(x - px)**2 + norm(y - py)**2)
    return 0.5 * (x - px) @ (beta * Ix + Q) @ (x - px) \
        + 0.5 * (y-py)@(beta*Iy+S) @ (y-py)


def adaptive_steps_residual_balance(primal_residual, dual_residual,
                                    smgap, tau, sigma,
                                    memory, eta=0.95):
    Delta = 1.5
    if len(memory) == 0:
        alpha = 0.5
    else:
        alpha = memory[-1][2]
        if alpha < 0.05:
            if alpha > 0.005:
                print('ici on arrete le warm up')
            alpha = 0

    up_down = 0
    if primal_residual > Delta * dual_residual:
        up_down = 1
        alpha = alpha * eta
    elif dual_residual > Delta * primal_residual:
        up_down = -1
        alpha = alpha * eta
    
    sigma = sigma * (1-alpha)**up_down
    tau = tau / (1-alpha)**up_down
    memory.append([tau, sigma, alpha, primal_residual, dual_residual])
    return tau, sigma

"""
not useful and does not seem to be working
def estimate_periodic_rate(instant_rates):
    # hypothese : log(rho(k)) = a cos(k theta + phi) + e(k)  où e(k) est iid
    logrho = np.log(instant_rates)

    Kmk = np.arange(len(logrho)-1, -1, -1)  # old samples are more difficult to fit
    def f(x):
        residuals = x[0] + x[1] * np.cos(x[2] * Kmk + x[3]) - logrho
        return (0.5/len(logrho) * norm(residuals)**2
                , 1/len(logrho) * (residuals) @ \
            np.vstack([np.ones(len(Kmk)),
                       np.cos(x[2] * Kmk + x[3]),
                       -x[1] * Kmk * np.sin(x[2] * Kmk + x[3]),
                       -x[1] * np.sin(x[2] * Kmk + x[3])]).T
                )
                 

    x0 = np.array([np.mean(logrho), 0.1, 0, 0.01])
    res = scipy.optimize.minimize(f, x0, jac=True, method='BFGS')
"""


def estimate_rate_iid(instant_rates):
    log_rho = np.log(instant_rates)
    if len(instant_rates) > 1:
        mu = np.mean(log_rho)
        std = np.std(np.log(instant_rates)) / np.sqrt(len(instant_rates))
    else:
        mu = 0
        std = 1000
    return mu, std

def estimate_rate_ar(instant_rates):
    log_rho = np.log(instant_rates)
    ar_order = 1
    if ar_order == 2:
        if len(instant_rates) > 3:
            A = np.hstack([np.ones((len(instant_rates)-2,1)), log_rho[1:-1].reshape((-1,1)), log_rho[0:-2].reshape(-1,1)])
            b = log_rho[2:]
            a = np.linalg.lstsq(A, b)[0]
            sigma = np.linalg.norm(A @ a - b) / np.sqrt(len(b))
            var = (1-a[2]) * sigma**2 / (1+a[2])*(1-a[1]-a[2])*(1+a[1]-a[2])
            if var < 0:
                var = 1e6
            return a[0] / (1 - a[1] - a[2]), np.sqrt(var) / np.sqrt(len(b))
        else:
            return np.mean(log_rho), 1000
    if ar_order == 1:
        if len(instant_rates) > 2:
            A = np.hstack([np.ones((len(instant_rates)-1,1)), log_rho[0:-1].reshape(-1,1)])
            b = log_rho[1:]
            a = np.linalg.lstsq(A, b)[0]
            sigma = np.linalg.norm(A @ a - b) / np.sqrt(len(b))
            var = sigma**2 / (1-a[1])**2            
            if np.abs(a[1]) >= 1:
                var = 1e6
            # print('ols=', a[0], a[1])
            return a[0] / (1 - a[1]), np.sqrt(var) / np.sqrt(len(b))
        else:
            return np.mean(log_rho), 1000
        
    

def adaptive_steps_convergence_monitoring(primal_residual, dual_residual,
                                          smgap, tau, sigma,
                                          memory,
                                          goldstein_warm_up=False,
                                          estimate_rate=estimate_rate_iid):

    if smgap <= 0:
        return tau, sigma

    alpha = 2  # having a power of 2 helps for the keys of the history list
    delta = 0.4

    log_rho = np.nan
    std = np.nan
    if len(memory) == 0:
        history = {tau[0]: []}
        updown = 1
        counter = 0
        reset_done = 0
        if goldstein_warm_up == True:
            alpha_rb = 0.5
        else:
            alpha_rb = 0
    else:
        adaptive_steps_residual_balance(primal_residual, dual_residual,
                                        smgap, tau, sigma,
                                        memory, eta=0.93)
        tau, sigma, alpha_rb, _, _ = memory.pop()
        history = memory[-1][0]
        counter = memory[-1][5]
        updown = memory[-1][3]
        reset_done = memory[-1][8]
        if counter == len(memory):
            counter_previous = 0
        else:
            counter_previous = memory[-counter-1][5]

        if len(memory) > 1:
            sm_gap_m1 = memory[-1][4]
            instant_rate = smgap / sm_gap_m1
            if tau[0] in history.keys():
                history[tau[0]].append(instant_rate)
                if len(history[tau[0]]) > len(memory) // 4:
                    history[tau[0]].pop(0)
            else:
                history[tau[0]] = [instant_rate]
            
            instant_rates = np.array(history[tau[0]])
            
            log_rho, std = estimate_rate(instant_rates)
            rho = np.exp(log_rho)

        if counter > max(3, counter_previous / 2):
            sm_gap_previous = memory[-counter][4]
            if counter_previous > 0:
                tau_previous = memory[-counter-1][1][0][0]
                instant_rates_previous = np.array(history[tau_previous])

                log_rho_previous, std_previous = estimate_rate(instant_rates_previous)
                std_next = np.sqrt(4 * alpha**2 * std**2 + alpha**4 * std_previous**2)

                rho_previous = np.exp(log_rho_previous)

                if rho < 1 and rho_previous < 1:
                    est_rho_next = 1 - np.exp(2 * np.log(1 - rho)
                                              - np.log(1 - rho_previous))

                    if tau[0] * alpha ** updown in history.keys():
                        tau_next = tau[0] * alpha ** updown
                        instant_rates_next = np.array(history[tau_next])
                        log_rho_next, std_next2 = estimate_rate(instant_rates_next)
                        if std_next2 < std_next:
                            est_rho_next = np.exp(log_rho_next)
                            std_next = std_next2
                else:
                    est_rho_next = 1

                # print("which is best?", 1-rho_previous, 1-rho, 1-est_rho_next)
            else:
                rho_previous = rho
                est_rho_next = 1
                std_previous = 1
                std_next = 1
            # rand = 0.5
            prob = 0.5 * sps.erfc((-rho_previous + est_rho_next)
                                             / np.sqrt(std_next**2 + std_previous**2) / np.sqrt(2))
            if smgap < delta * sm_gap_previous and rho < 1 and rho_previous < 1:
                if counter < len(memory):
                    if prob < 0.53:
                        # either worse rate or not enough information: revert gear
                        updown = -updown
                        # est_rho_next = rho_previous
                    
                    # print(rand,  0.5*sps.erfc((rho_previous - est_rho_next) / np.sqrt(4 * alpha**2 * std**2 + (1+alpha ** 4) * std_previous**2) / np.sqrt(2)), updown, np.sqrt(4 * alpha**2 * std**2 + (1+alpha ** 4) * std_previous**2))
                    # pdb.set_trace()
                    tau_ = tau * alpha ** updown
                    sigma_ = sigma * alpha ** (-updown)
                    print('changing step sizes', tau[0], '->', tau_[0], 'counter = ', counter, '1-est_rho_next = ', 1-est_rho_next, '1-rho = ', 1-rho, '1-rho_previous = ', 1-rho_previous)
                    print('prob=', prob, np.exp(2 * np.log(1 - rho)
                                                    - np.log(1 - rho_previous)), 1 - est_rho_next, std_next, np.sqrt(4 * alpha**2 * std**2 + alpha**4 * std_previous**2))
                    tau = tau_
                    sigma = sigma_
                counter = 0
            # elif reset_done == 0 and smgap < delta**2 * sm_gap_previous and counter > 1000:
            #    print('rate estimate does not converge, reset counter once')
            #    reset_done = 1
            #    counter = 0
                
    counter += 1
    memory.append([history, [tau, sigma], alpha_rb, updown, smgap, counter, log_rho, std, reset_done])
    return tau, sigma


def adaptive_steps_convergence_monitoring_v3(primal_residual, dual_residual,
                                          smgap, tau, sigma,
                                          memory,
                                          goldstein_warm_up=True,
                                          estimate_rate=estimate_rate_iid):

    if smgap <= 0:
        return tau, sigma

    alpha = 2  # having a power of 2 helps for the keys of the history list
    delta = 0.4
    delta_warmup = 0.01

    log_rho = np.nan
    std = np.nan
    if len(memory) == 0:
        history = {sigma[0]: []}
        updown = 1
        counter = 0
        reset_done = 0
        sigma_checked = sigma[0]
        if goldstein_warm_up == True:
            alpha_rb = 0.5
            sigma_balance = sigma[0]
            do_residual_balance = True
            sigma_tentative = None
        else:
            alpha_rb = 0
            sigma_balance = None
            do_residual_balance = False
            sigma_tentative = sigma_checked * (alpha ** updown)
    else:
        sigma_checked, sigma_balance, sigma_tentative = memory[-1][8]
        do_residual_balance = memory[-1][9]
        if do_residual_balance == True:
            adaptive_steps_residual_balance(primal_residual, dual_residual,
                                            smgap, tau, sigma,
                                            memory)
            tau, sigma, alpha_rb, _, _ = memory.pop()
            sigma_balance = sigma[0]
        else:
            alpha_rb = memory[-1][2]
            
        history = memory[-1][0]
        counter = memory[-1][5]
        updown = memory[-1][3]
        if counter == len(memory):
            counter_previous = 0
        else:
            counter_previous = memory[-counter-1][5]

        if len(memory) > 1:
            sm_gap_m1 = memory[-1][4]
            instant_rate = smgap / sm_gap_m1
            if sigma[0] in history.keys():
                history[sigma[0]].append(instant_rate)
                if len(history[sigma[0]]) > len(memory) // 4:
                    history[sigma[0]].pop(0)
            else:
                history[sigma[0]] = [instant_rate]
            
            instant_rates = np.array(history[sigma[0]])
            
            log_rho, std = estimate_rate(instant_rates)
            rho = np.exp(log_rho)

        if counter > 3:
            sm_gap_previous = memory[-counter][4]

            if do_residual_balance == True:
                if smgap < delta_warmup * sm_gap_previous:
                    do_residual_balance = False
                    if sigma_balance > sigma_checked:
                        updown = 1
                        if sigma_balance < sigma_checked * alpha ** updown:
                            sigma_tentative = sigma_checked * alpha ** updown
                        else:
                            sigma_tentative = sigma_balance
                    if sigma_balance <= sigma_checked:
                        updown = -1
                        if sigma_balance > sigma_checked * alpha ** updown:
                            sigma_tentative = sigma_checked * alpha ** updown
                        else:
                            sigma_tentative = sigma_balance
                        
                    print('goldstein warm up finished', sigma_checked, sigma_balance, sigma_tentative)

            else:

                instant_rates_checked = np.array(history[sigma_checked])
                log_rho_checked, std_checked = estimate_rate(instant_rates_checked)

                if sigma_tentative in history.keys():
                    instant_rates_tentative = np.array(history[sigma_tentative])
                    log_rho_tentative, std_tentative = estimate_rate(instant_rates_tentative)
                else:
                    log_rho_tentative = log_rho_checked
                    std_tentative = 1000

                prob = 0.5 * sps.erfc((-log_rho_checked + log_rho_tentative)
                                             / np.sqrt(std_checked**2 + std_tentative**2) / np.sqrt(2))
                if smgap < delta * sm_gap_previous:
                    if counter < len(memory):
                        if prob <= 0.55 and prob >= 0.45:
                            # not enough information: try to increase knowledge
                            if sigma[0] == sigma_tentative:
                                tau_ = sigma[0] / sigma_checked * tau
                                sigma_ = sigma_checked / sigma[0] * sigma
                            else:  # sigma[0] = sigma_checked
                                tau_ = sigma[0] / sigma_tentative * tau
                                sigma_ = sigma_tentative / sigma[0] * sigma
                        elif prob < 0.45:
                            # the tentative is a failure
                            if sigma_tentative == sigma_balance:
                                # we forget about trying to do residual balance
                                updown = -updown
                                sigma_tentative = sigma_checked * (alpha ** updown)
                            else:
                                updown = -updown
                                sigma_tentative = sigma_checked * (alpha ** updown)
                            if sigma[0] == sigma_checked:
                                tau_ = sigma[0] / sigma_tentative * tau
                                sigma_ = sigma_tentative / sigma[0] * sigma
                            else:
                                tau_ = sigma[0] / sigma_checked * tau
                                sigma_ = sigma_checked / sigma[0] * sigma
                                
                        else:  # prob > 0.55:
                            # the tentative is a success
                            sigma_checked = sigma_tentative
                            sigma_tentative = sigma_checked * (alpha ** updown)
                            tau_ = sigma[0] / sigma_tentative * tau
                            sigma_ = sigma_tentative / sigma[0] * sigma
                            
                        print('prob=', prob, std_checked, std_tentative)
                        if prob > 0.45 and prob < 0.55:
                            print('changing step sizes', sigma[0], '->', sigma_[0], 'counter = ', counter)
                        else:
                            print('changing step sizes', sigma[0], '->', sigma_[0], 'counter = ', counter, '1-est_rho_tentative = ', 1-np.exp(log_rho_tentative), '1-rho_checked = ', 1-np.exp(log_rho_checked), sigma_checked, sigma_tentative)
                        tau = tau_
                        sigma = sigma_
                    counter = 0
                
    counter += 1
    memory.append([history, [tau, sigma], alpha_rb, updown, smgap, counter, log_rho, std, [sigma_checked, sigma_balance, sigma_tentative], do_residual_balance])
    return tau, sigma



def pure_cd(x_init, y_init, sigma0, K, adaptive_steps=None, verbose=True):
    x = x_init.copy()
    y = y_init.copy()
    x_ = x_init.copy()
    y_ = y_init.copy()
    Nx = len(x)
    Ny = len(y)
    z_star = np.linalg.solve(M1, np.hstack([-c, b]))
    x_star = z_star[:Nx]
    y_star = z_star[Nx:]

    J = []
    for i in range(Nx):
        J.append(np.nonzero(A.T[i])[0])
    theta = np.zeros(Ny)
    for j in range(Ny):
        theta[j] = len(np.nonzero(A[j])[0])

    sigma = sigma0
    tau = gamma / np.sum(((theta * sigma) * A.T**2).T, axis=0)
    
    NX = np.zeros(K//verbose)
    NY = np.zeros(K//verbose)
    DG = np.zeros(K//verbose)
    memory = []
    rx = A @ x
    estimate_subgradient_primal = 0
    estimate_subgradient_dual = 0
    length_of_estimate = 0
    
    for k in range(K):
        i = np.random.randint(0, Nx)
        j = J[i]
        y_[j] = 1 / (1+sigma[j]*S[0,0]) *(y[j] + sigma[j] * (rx[j] - b[j]))
        x_[i] = 1 / (1+tau[i]*Q[0,0]) * (x[i] - tau[i] * (A.T[i] @ y_ + c[i]))

        estimate_subgradient_primal += ((x_[i] - x[i]) / tau[i])**2 * Nx # the expectation of the squared norm of this is what we want
        estimate_subgradient_dual += norm((y_[j] - y[j])/sigma[j] * Nx / len(j) + (theta[j]-1) * (A[j,i] * (Nx * (x_[i] - x[i]))), 2)**2 / (Nx / len(j))
        length_of_estimate += 1

        y[j] = y_[j] + (sigma[j] * theta[j]) * (A[j,i] * (x_[i] - x[i]))
        rx = rx + A[:,i] * (x_[i] - x[i])
        x[i] = x_[i]

        if adaptive_steps is not None:
            if length_of_estimate == Nx:
                primal_residual = estimate_subgradient_primal / Nx
                dual_residual = estimate_subgradient_dual / Nx
                length_of_estimate = 0
                estimate_subgradient_primal = 0
                estimate_subgradient_dual = 0

                smgap = smoothed_gap(x, y, beta=1)
                residual_norm = primal_residual + dual_residual
                
                tau, sigma = adaptive_steps(primal_residual,
                                            dual_residual, smgap,
                                            tau, sigma,
                                            memory)

        if k % verbose == 0:
            NX[k//verbose] = norm(x - x_star)
            NY[k//verbose] = norm(y - y_star)
            DG[k//verbose] = smoothed_gap(x, y, 1e-30)

            if DG[k//verbose] < 1e-10:
                print(k)
                break

            if NX[k//verbose] + NY[k//verbose] < 1e-14:
                break

    return x, y, NX, NY, DG, memory
        

def rapdhg(x_init, y_init, tau, sigma, period, K):
    x = x_init.copy()
    y = y_init.copy()

    x_star = spl.spsolve(A, b)
    y_star = spl.spsolve(A.T, -c)
    
    Nx = np.zeros(K)
    Ny = np.zeros(K)
    ktot = 0

    while ktot < K:
        x_in, y_in, avgx, avgy, Nx_in, Ny_in, _ = pdhg(x, y,  tau, sigma,
                                                    min(period, K-ktot))
        x = avgx
        y = avgy
        # x = x - tau * period * (A.T @ avgy + c)  # stable mais lent
        # y = y + sigma * period * (A @ avgx - b)
        Nx[ktot:ktot+len(Nx_in)] = Nx_in
        Ny[ktot:ktot+len(Ny_in)] = Ny_in
        ktot = ktot + period

    return x, y, x_in, y_in, Nprimal, Ndual


#TAUS = 0.001 * 2 ** np.arange(20)
#z0 = np.arange(Nx+Ny)
#Goldstein = [primal_dual_residuals_ratio(tau, z0, K=1)[-1] for tau in TAUS]
#Deriv_eig = [deriv_eigenvalue(tau)[-1] for tau in TAUS]
#Deriv_res = [deriv_squared_residual(tau, z0) for tau in TAUS]

z0 = np.ones(Nx+Ny)

theta = np.zeros(Ny)
for j in range(Ny):
    theta[j] = len(np.nonzero(A[j]))

normA = np.sum(A**2, axis=0)

sigma0 = 10 / theta / max(normA)
# sigma* = 0.01 / theta / max(normA)

K = 150000

x_cm, y_cm, NX_cm, NY_cm, DG_cm, memory_cm = pure_cd(z0[:Nx], z0[Nx:], sigma0, K, adaptive_steps=adaptive_steps_convergence_monitoring_v3, verbose=1)

def adaptive_steps_convergence_monitoring_v3_ar_no_warmup(primal_residual, dual_residual,
                                          smgap, tau, sigma,
                                          memory):
    return adaptive_steps_convergence_monitoring_v3(primal_residual, dual_residual,
                                          smgap, tau, sigma,
                                                    memory,
                                          goldstein_warm_up=False,
                                          estimate_rate=estimate_rate_ar)

def adaptive_steps_convergence_monitoring_v3_no_warmup(primal_residual, dual_residual,
                                          smgap, tau, sigma,
                                          memory):
    return adaptive_steps_convergence_monitoring_v3(primal_residual, dual_residual,
                                          smgap, tau, sigma,
                                                    memory,
                                          goldstein_warm_up=False,
                                                    estimate_rate=estimate_rate_iid)

x_cm2, y_cm2, NX_cm2, NY_cm2, DG_cm2, memory_cm2 = pure_cd(z0[:Nx], z0[Nx:], sigma0, K, adaptive_steps=adaptive_steps_convergence_monitoring_v3_ar_no_warmup, verbose=1)
x_cm3, y_cm3, NX_cm3, NY_cm3, DG_cm3, memory_cm3 = pure_cd(z0[:Nx], z0[Nx:], sigma0, K, adaptive_steps=adaptive_steps_convergence_monitoring_v3_no_warmup, verbose=1)

x_rb, y_rb, NX_rb, NY_rb, DG_rb, memory_rb = pure_cd(z0[:Nx], z0[Nx:], sigma0, K, adaptive_steps=adaptive_steps_residual_balance, verbose=1)

x, y, NX, NY, DG, memory = pure_cd(z0[:Nx], z0[Nx:], sigma0, K, adaptive_steps=None, verbose=1)


plt.figure(1)
plt.semilogy(np.vstack([NX, NY]).T)
plt.semilogy(np.vstack([NX_rb, NY_rb]).T, '--')
plt.semilogy(np.vstack([NX_cm, NY_cm]).T, '-.')
plt.legend(['NX', 'NY', 'NX_rb', 'Ny_rb', 'NX_cm', 'NY_cm'])
plt.show()

plt.figure(2)
plt.semilogy(np.array([DG, DG_rb, DG_cm, DG_cm2, DG_cm3]).T)
plt.legend(['pure-cd', 'pure-cd + residual balance', 'pure-cd + rb + cm idd', 'puce-cd + cm ar', 'pure-cd + cm iid'])
plt.ylabel('Duality gap')
plt.xlabel('iteration')
plt.show()

SIGMA_rb = np.array([memory_rb[k][1][0] for k in range(len(memory_rb))])
SIGMA_cm = np.array([memory_cm[k][1][1][0] for k in range(len(memory_cm))])

plt.figure(3)
plt.semilogy(sigma0[0] * np.ones(len(SIGMA_cm)), '-')
plt.semilogy(SIGMA_rb, '--')
plt.semilogy(SIGMA_cm, '-.')
plt.show()


SMGAPS = np.array([memory_cm[k][4] for k in range(len(memory_cm))])
instant_rates = SMGAPS[1:] / SMGAPS[:-1]
plt.figure(4)
#plt.plot(instant_rates)
#N = 100
#plt.plot(np.convolve(instant_rates, 1/N * np.ones(N)))
RATES = np.array([memory_cm[k][6] for k in range(len(memory_cm))])
STD = np.minimum(1, np.array([memory_cm[k][7] for k in range(len(memory_cm))]))
plt.plot(RATES)
plt.plot(RATES + 0.5*STD)
plt.plot(RATES - 0.5*STD)
plt.show()

"""
# Ça m'a permis de détecter un bug.
time_for_conv = []
SIGMA = 2. ** np.arange(3, -10, -1)
DDGG = []
for sigma in SIGMA:
    sigma = sigma / theta / max(normA)
    x, y, NX, NY, DG, memory = pure_cd(z0[:Nx], z0[Nx:], sigma, K, adaptive_steps=None, verbose=100)
    DDGG.append(DG)
    converged = np.nonzero(DG<1e-8)[0]
    if len(converged) == 0:
        time_for_conv.append(len(DG))
    else:
        time_for_conv.append(np.min(converged))
    plt.semilogy(DG)
    plt.show()
    print(sigma[0], time_for_conv[-1])
    gradletters@berkeley.edu
plt.figure(4)
plt.semilogx(SIGMA/theta[0]/norm(A), np.array(time_for_conv) * 100)
plt.xlabel('$\sigma$')
plt.ylabel('iterations before convergence')
plt.tight_layout()
plt.show()

sigma* = 5e-3
"""


if 0:
    val = []
    for ss in np.arange(0.01, 1, 0.01):
        # tests sur les valeurs propres de la matrice de point fixe   
        Ix = np.eye(Nx)
        Iy = np.eye(Ny)
        I = np.eye(Nx + Ny)

        J = []
        for i in range(Nx):
            J.append(np.nonzero(A.T[i])[0])
        theta = np.zeros(Ny)
        for j in range(Ny):
            theta[j] = len(np.nonzero(A[j])[0])

        sigma = ss / theta / norm(A)
        tau = gamma / np.sum(((theta * sigma) * A.T**2).T, axis=0)

        Rbar = np.vstack([np.hstack([inv(Ix+tau*Q) @ (Ix - np.diag(tau) @ A.T @ (inv(Iy+sigma*S) * sigma) @ A),
                              - (tau * inv(Ix+tau*Q)) @ A.T @ inv(Iy+sigma*S)]),
                   np.hstack([(sigma * inv(Iy+sigma*S)) @ A,
                              inv(Iy+sigma*S)])])

        Rb = np.hstack([-(tau*inv(Ix+tau*Q)) @ (- A.T @ inv(Iy+sigma*S) @ (sigma * b) + c),
                        -sigma*inv(Iy+sigma*S) @ b])


        Sk = []
        for i in range(Nx):
            Sk.append(np.zeros((Nx+Ny, Nx+Ny)))
            Sk[i][i,i] = 1
            Sk[i][Nx+J[i], i] = np.diag(sigma[J[i]] * theta[J[i]]) @ A[J[i],i]
            for j in J[i]:
                Sk[i][Nx+j, Nx+j] = 1

            # z = z + S @ (Rbar @ z + Rb - z)
            # Rbar @ zstar + Rb = 0

        RR = I
        for i in range(Nx):
            RR = (I + Sk[i] @ (Rbar - I)) @ RR

        RRr = I
        for ii in range(10*Nx):
            i = np.random.randint(Nx)
            RRr = (I + Sk[i] @ (Rbar - I)) @ RRr
        val.append(min(1-np.abs(np.linalg.eig(RRr)[0])**(1/10)))
        
